/*
* HTC Kaiser Keyboard based on:
*			Palm TC hardware buttons (Holger Bocklet, Bitz.Email@gmx.net, )
*
* Author: Marcin Baliniak <marcin@baliniak.pl>
*
* Tips:
* WinMenu key = Ctrl
* Camera = Backslash
* Camera + Shift = <Pipe>
* PAD Center = KEY_OK
* PAD OK = KEY_BACK			//<- we don need 2 OK buttons, OK remaped to BACK, PAD_CENTER = OK
*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/jiffies.h>
#include <linux/platform_device.h>
#include <linux/workqueue.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/sched.h>
#include <linux/irq.h>

#include <linux/gpiodev2.h>
#include <asm/gpio.h>
#include <asm/arch/board.h>

#include <asm/mach/arch.h>
#include <asm/mach-types.h>
#include <asm/arch/irqs.h>

//#define USE_RELEASE_TIMER

#define MODULE_NAME "kaiser_kbd"

//#define KAISER_KEYBOARD_DEBUG

#ifdef KAISER_KEYBOARD_DEBUG
#define DBG(x...) \
	printk(KERN_INFO MODULE_NAME ": " x)
#else
#define DBG(x...)	do { } while (0)
#endif

#define FALSE		0
#define TRUE		1
#define PERMANENT	2

#define PRESSED_BIT	0x01
#define ALT_BIT		0x02
#define CAPS_BIT	0x04
#define CTRL_BIT	0x08
#define SHIFT_BIT	0x80

#define KEY_ANDROID_OK 		232
#define KEY_ANDROID_MENU	229

#ifdef USE_RELEASE_TIMER
static void     release_timer_went_off (unsigned long);
static struct timer_list key_release_timer;
#endif

static u8	alternate_key = FALSE;
static u8	shift_key = FALSE;
static u8	ctrl_key = FALSE;
static u8	col_gpio[] = { 36, 37, 38, 39, 40, 41, 42 };
static u8	row_gpio[] = {23, 30, 31, 32, 33, 34, 35, 78};

#define RELEASE_CHECK_TIME_MS	50
#define MAX_ROW	ARRAY_SIZE(row_gpio)
#define MAX_COL	ARRAY_SIZE(col_gpio)
#define KEYBOARD_TILT (KEY_MAX+1)
#define BACKLIGHT_TIMEOUT (15*20) //15s

static struct {
	struct input_dev *idev;
	struct workqueue_struct *workqueue;
	struct work_struct task;
	struct semaphore sem;
	struct platform_device *platform_dev;
	struct timer_list timer20Hz;		// timer 20Hz to read gpio changes
	spinlock_t lock;// = SPIN_LOCK_UNLOCKED;
	unsigned int backlight_timer;
} kbd_data;



static struct //gpio buttons
{
	unsigned short 	key;		//key
	unsigned short 	alt_key;		//alternative key
	unsigned short	gpio;		//gpio num
	unsigned char	state;	//to remember old state
	char 			*desc;
} gpio_buttons[]={
	//{KEY_VOLUMEUP, 18, 0, "Vol UP"},
	//{KEY_VOLUMEDOWN, 19, 0, "Vol Down"},
	{KEY_POWER, KEY_POWER, 20, 0, "Power"},
	{KEY_UP, KEY_UP, EGPIO_5_0_PAD_UP, 0, "PAD UP"},
	{KEY_DOWN, KEY_DOWN, EGPIO_5_1_PAD_DOWN, 0, "PAD DOWN"},
	{KEY_LEFT, KEY_LEFT, EGPIO_5_2_PAD_LEFT, 0, "PAD LEFT"},
	{KEY_RIGHT, KEY_RIGHT, EGPIO_5_3_PAD_RIGHT, 0, "PAD RIGHT"},
	{KEY_ANDROID_OK, KEY_ANDROID_OK, EGPIO_5_4_PAD_CENTER, 0, "PAD CENTER"},

	{KEY_UP, KEY_PAGEUP, EGPIO_6_0_KBD_UP, 0, "KBD UP/PAGE UP"},
	{KEY_DOWN, KEY_PAGEDOWN, EGPIO_6_1_KBD_DOWN, 0, "KBD DOWN/PAGE DOWN"},
	{KEY_LEFT, KEY_COMMA, EGPIO_6_2_KBD_LEFT, 0, "KBD LEFT/COMMA"},
        {KEY_RIGHT, KEY_DOCUMENTS, EGPIO_6_3_KBD_RIGHT, 0, "KBD RIGHT/DOC."},

	{KEY_BACKSLASH, KEY_CAMERA, EGPIO_6_4_KEY_CAM0, 0, "BACKSLASH/Camera 0"},
	{KEY_CAMERA, KEY_CAMERA, EGPIO_6_5_KEY_CAM1, 0, "Camera 1"},
	
 {KEYBOARD_TILT, KEYBOARD_TILT, EGPIO_5_5_TILT, 0, "Tilt"},
};

static struct
{																// matrix for real buttons
	unsigned short  key;
	unsigned short  alt_key;
	u8              flags;				// for recording last status (lowest) and shift_bit (highest)
	char           *desc;
} matrix_buttons[MAX_ROW][MAX_COL] = {
//0
	{
		{KEY_LEFTCTRL, KEY_SEMICOLON, FALSE, "Menu/;"}, //<- I repleace this key with ctrl
		{KEY_ANDROID_OK, KEY_WLAN, FALSE, "OK/WLAN"},
		{KEY_TAB, KEY_TAB, FALSE, "TAB"},
		{KEY_V, KEY_8, SHIFT_BIT, "V/*"},
		{KEY_SPACE, KEY_SPACE, FALSE, "SPACE"},
		{KEY_DOT, KEY_0, FALSE, "DOT/0"},
		{KEY_ENTER, KEY_ENTER, FALSE, "ENTER"}
	},
//1
	{
		{KEY_MAX, KEY_MAX, FALSE, "none"},
		{KEY_MAX, KEY_MAX, FALSE, "none"},
		{KEY_MAX, KEY_MAX, FALSE, "none"},
		{KEY_MENU, KEY_MENU, FALSE, "Pad Menu"},
		{KEY_END, KEY_END, FALSE, "End"}, //red
		{KEY_BACK, KEY_BACK, FALSE, "Pad OK"},//we don't have Back or C key ,so I remaped OK 
		{KEY_SEND, KEY_SEND, FALSE, "Send"} //call
	},
//2
	{
		{KEY_FN, KEY_FN, FALSE, "Fn"},
		{KEY_X, KEY_7, SHIFT_BIT, "X/&"},
		{KEY_C, KEY_SEMICOLON, SHIFT_BIT, "C/:"},
		{KEY_G, KEY_EQUAL, SHIFT_BIT, "G/+"},
		{KEY_B, KEY_7, FALSE, "B/7"},
		{KEY_M, KEY_9, FALSE, "M/9"},
		{KEY_BACKSPACE, KEY_DELETE, FALSE, "Back./Del"}
	},
//3
	{
		{KEY_RIGHTSHIFT, KEY_RIGHTSHIFT, FALSE, "Shift"},
		{KEY_Z, KEY_APOSTROPHE, FALSE, "Z/'"},
		{KEY_F, KEY_MINUS, FALSE, "F/-"},
		{KEY_Y, KEY_1, FALSE, "Y/1"},
		{KEY_H, KEY_4, FALSE, "H/4"},
		{KEY_N, KEY_8, FALSE, "N/8"},
		{KEY_L, KEY_EQUAL, FALSE, "L/="}
	},
//4
	{
		{KEY_A, KEY_MINUS, SHIFT_BIT, "A/_"},
		{KEY_S, KEY_APOSTROPHE, SHIFT_BIT, "S/\""},
		{KEY_D, KEY_5, SHIFT_BIT, "D/%"},
		{KEY_T, KEY_3, SHIFT_BIT, "T/#"},
		{KEY_J, KEY_5, FALSE, "J/5"},
		{KEY_K, KEY_6, FALSE, "K/6"},
		{KEY_P, KEY_0, SHIFT_BIT, "P/)"}
	},
//5
	{
		{KEY_Q, KEY_2, SHIFT_BIT, "Q/@"},
		{KEY_W, KEY_1, SHIFT_BIT, "W/!"},
		{KEY_E, KEY_SLASH, SHIFT_BIT, "E/?"},
		{KEY_R, KEY_SLASH, FALSE, "R//"},
		{KEY_U, KEY_2, FALSE, "U/2"},
		{KEY_I, KEY_3, FALSE, "I/3"},
		{KEY_O, KEY_9, SHIFT_BIT, "0/("}
	},
//6
	{
		{KEY_MAX, KEY_MAX, FALSE, "61"},
		{KEY_MAX, KEY_MAX, FALSE, "62"},
		{KEY_MAX, KEY_MAX, FALSE, "63"},
		{KEY_TAB, KEY_F7, FALSE, "Kbd LeftSoft"},
		{KEY_MAX, KEY_MAX, FALSE, "65"},
		{KEY_TAB, KEY_F7, FALSE, "Kbd RightSoft"},
		{KEY_MAX, KEY_MAX, FALSE, "67"}
	},
//7
	{
		{KEY_MENU, KEY_SEMICOLON, FALSE, "Left side OK"},
		{KEY_MAX, KEY_MAX, FALSE, "72"},
		{KEY_TAB, KEY_F7, FALSE, "VoiceMemo"},
		{KEY_TAB, KEY_F7, FALSE, "Pad IE"},
		{KEY_TAB, KEY_F7, FALSE, "Pad RightSoft"},
		{KEY_TAB, KEY_F7, FALSE, "Pad Mail"},
		{KEY_V, KEY_F8, FALSE, "Pad LeftSoft"}
	},
};

static void kaiser_kbd_enable_irqs(void)
{
	int i;
		for (i=0;i<ARRAY_SIZE(col_gpio); i++)
			enable_irq(gpio_to_irq(col_gpio[i]));
		//Jog wheel
		enable_irq(gpio_to_irq(18));
		enable_irq(gpio_to_irq(19));
}

static void kaiser_kbd_disable_irqs(void)
{
	int i;
		for (i=0;i<ARRAY_SIZE(col_gpio); i++)
			disable_irq(gpio_to_irq(col_gpio[i]));
		//Jog wheel
		disable_irq(gpio_to_irq(18));
		disable_irq(gpio_to_irq(19));
}

static irqreturn_t kaiser_kbd_irq_handler (int irq, void *dev_id)
{
	int i;
	unsigned long   flags;
	
	kaiser_kbd_disable_irqs();
	
	DBG("kaiser_kbd_irq_handler()\n");
		
	if (gpio_get_value(18) == 1 && gpio_get_value(19)==0)
	{
		DBG("JOG_UP\n");
		input_report_key(kbd_data.idev, KEY_UP, 1);
		input_report_key(kbd_data.idev, KEY_UP, 0);
		input_sync(kbd_data.idev);
		kaiser_kbd_enable_irqs();
		return IRQ_HANDLED;
	}
	if (gpio_get_value(18) == 0 && gpio_get_value(19)==1)
	{
		DBG("JOG_DOWN\n");	
		input_report_key(kbd_data.idev, KEY_DOWN, 1);
		input_report_key(kbd_data.idev, KEY_DOWN, 0);
		input_sync(kbd_data.idev);
		kaiser_kbd_enable_irqs();
		return IRQ_HANDLED;
	}
	// set row-gpios high so we can pull them low later one-by-one in loop
	spin_lock_irqsave (&kbd_data.lock, flags);
	for (i = 0; i < MAX_ROW; i++)
		gpio_set_value (row_gpio[i], 1);

	spin_unlock_irqrestore (&kbd_data.lock, flags);

	queue_work (kbd_data.workqueue, &kbd_data.task);
	return IRQ_HANDLED;
}

void kaiser_update_leds(void)
{
	if (ctrl_key)
	{
		//booth leds indicates ctrl pressed
		gpio_set_value(EGPIO_3_2_FN_LED,1);
		gpio_set_value(EGPIO_3_3_CAPS_LED,1);
	}
	else
	{
		if (alternate_key)
			gpio_set_value(EGPIO_3_2_FN_LED,1);
		else
			gpio_set_value(EGPIO_3_2_FN_LED,0);

		if (shift_key)
			gpio_set_value(EGPIO_3_3_CAPS_LED,1);
		else
			gpio_set_value(EGPIO_3_3_CAPS_LED,0);
	}
}

static void kaiser_kbd_queuework (struct work_struct *data)
{
	unsigned long   flags;
	int             gpio, row, col;
	// run thru rows, pull row-gpio low and see if what gpio of the colums goes low with it
	for (row = 0; row < MAX_ROW; row++)
	{
		spin_lock_irqsave (&kbd_data.lock, flags);
		gpio_set_value (row_gpio[row], 0);
		spin_unlock_irqrestore (&kbd_data.lock, flags);

		udelay(500);

		for (col = 0; col < MAX_COL; col++)
		{
			spin_lock_irqsave (&kbd_data.lock, flags);
			gpio = gpio_get_value (col_gpio[col]);
			spin_unlock_irqrestore (&kbd_data.lock, flags);

			if (!gpio)
			{
				if (!(matrix_buttons[row][col].flags & PRESSED_BIT))
				{  //is not pressed at the moment
					kbd_data.backlight_timer = BACKLIGHT_TIMEOUT;
					gpio_set_value(EGPIO_1_2_KBD_BAKLIGHT,1);
					matrix_buttons[row][col].flags |= PRESSED_BIT;
					if (matrix_buttons[row][col].key == KEY_FN)
					{
						alternate_key++;
						if (alternate_key>2)
							alternate_key=0;
						shift_key=0;
						ctrl_key=0;
					}
					else if (matrix_buttons[row][col].key == KEY_RIGHTSHIFT)
					{
						shift_key++;
						if (shift_key>2)
							shift_key=0;
						alternate_key=0;
						ctrl_key=0;
					}
					else if (matrix_buttons[row][col].key == KEY_LEFTCTRL && alternate_key == 0)
					{
						ctrl_key++;
						if (ctrl_key>2)
							ctrl_key=0;
						shift_key=0;
					}
					else
					{
						if (alternate_key)
						{
							if (matrix_buttons[row][col].alt_key < KEY_MAX)
							{
								matrix_buttons[row][col].flags |= ALT_BIT;
								if (matrix_buttons[row][col].flags & SHIFT_BIT)
									input_report_key (kbd_data.idev, KEY_RIGHTSHIFT, 1);
								input_report_key (kbd_data.idev, matrix_buttons[row][col].alt_key, 1);
								input_sync (kbd_data.idev);
							}
							if (alternate_key != PERMANENT)
								alternate_key = FALSE;
						}
						else if (shift_key)
						{
							if (matrix_buttons[row][col].key < KEY_MAX)
							{
								matrix_buttons[row][col].flags |= CAPS_BIT;
								input_report_key (kbd_data.idev, KEY_RIGHTSHIFT, 1);
								input_report_key (kbd_data.idev, matrix_buttons[row][col].key, 1);
								input_sync (kbd_data.idev);
							}
							if (shift_key != PERMANENT)
								shift_key = FALSE;
						}
						else if (ctrl_key)
						{
							if (matrix_buttons[row][col].key < KEY_MAX)
							{
								matrix_buttons[row][col].flags |= CTRL_BIT;
								input_report_key (kbd_data.idev, KEY_LEFTCTRL, 1);
								input_report_key (kbd_data.idev, matrix_buttons[row][col].key, 1);
								input_sync (kbd_data.idev);
							}
							if (ctrl_key != PERMANENT)
								ctrl_key = FALSE;
						}
						else
						{
							if (matrix_buttons[row][col].key < KEY_MAX)
							{
								input_report_key (kbd_data.idev, matrix_buttons[row][col].key, 1);
								input_sync (kbd_data.idev);
							}
						}
					}
					kaiser_update_leds();
				}
#ifdef USE_RELEASE_TIMER
				spin_lock_irqsave (&kbd_data.lock, flags);
				if (timer_pending (&key_release_timer))
				{
					mod_timer (&key_release_timer, jiffies + msecs_to_jiffies (RELEASE_CHECK_TIME_MS));
				}
				else
				{
					key_release_timer.expires = (jiffies + msecs_to_jiffies (RELEASE_CHECK_TIME_MS));
					add_timer (&key_release_timer);
				}
				spin_unlock_irqrestore (&kbd_data.lock, flags);
#endif
			}
			else
			{	// not pressed branch
				if (matrix_buttons[row][col].flags & PRESSED_BIT)
				{
					matrix_buttons[row][col].flags ^= PRESSED_BIT;	//xor bit to 0
					if (matrix_buttons[row][col].key < KEY_MAX)
					{
						//alt key release
						if (matrix_buttons[row][col].flags & ALT_BIT)
						{
							matrix_buttons[row][col].flags ^= ALT_BIT;	//xor bit to 0
							input_report_key (kbd_data.idev, matrix_buttons[row][col].alt_key, 0);
							if (matrix_buttons[row][col].flags & SHIFT_BIT)
								input_report_key (kbd_data.idev, KEY_RIGHTSHIFT, 0);
						}
						else
						{
							input_report_key (kbd_data.idev, matrix_buttons[row][col].key, 0);
							if (matrix_buttons[row][col].flags & CAPS_BIT)
							{
								matrix_buttons[row][col].flags ^= CAPS_BIT;	//xor bit to 0
								input_report_key (kbd_data.idev, KEY_RIGHTSHIFT, 0);
							}
							if (matrix_buttons[row][col].flags & CTRL_BIT)
							{
								matrix_buttons[row][col].flags ^= CTRL_BIT;	//xor bit to 0
								input_report_key (kbd_data.idev, KEY_LEFTCTRL, 0);
							}
						}
						input_sync (kbd_data.idev);
					}
				}
			}
		}
		// back to high for this row
		spin_lock_irqsave (&kbd_data.lock, flags);
		gpio_set_value (row_gpio[row], 1);
		spin_unlock_irqrestore (&kbd_data.lock, flags);
	}

	// set row-gpios low again
	spin_lock_irqsave (&kbd_data.lock, flags);
	for (row = 0; row < MAX_ROW; row++)
		gpio_set_value (row_gpio[row], 0);

	spin_unlock_irqrestore (&kbd_data.lock, flags);

	// restore irqs
	kaiser_kbd_enable_irqs();
}


#ifdef USE_RELEASE_TIMER
/*
* This is called when the key release timer goes off.  That's the
* only time it should be called.  Check for changes in what keys
* are down.
*/
static void release_timer_went_off (unsigned long unused)
{
	unsigned long   flags;

	spin_lock_irqsave (&kbd_data.lock, flags);

	queue_work (kbd_data.workqueue, &kbd_data.task);

	spin_unlock_irqrestore (&kbd_data.lock, flags);
}
#endif

//we need to read gpio buttons in pooling mode, no irq fo EGPIO
static void kbd_handle_timer20Hz(unsigned long __data)
{
	int i,state;
	unsigned long   flags;

	spin_lock_irqsave (&kbd_data.lock, flags);

	for (i = 0; i < ARRAY_SIZE(gpio_buttons); i++)
	{
		state = gpio_get_value(gpio_buttons[i].gpio);
		if (gpio_buttons[i].state != state)
		{
			gpio_buttons[i].state = state;
			DBG("New state of %s:%d\n",gpio_buttons[i].desc,state);
			if (!state)//key pressed
			{
				if (gpio_buttons[i].key == KEYBOARD_TILT)
					kbd_data.backlight_timer = 0;
				
				if (shift_key)
					input_report_key (kbd_data.idev, KEY_RIGHTSHIFT, 1);
				
				if (alternate_key)
				{
					input_report_key (kbd_data.idev, gpio_buttons[i].alt_key, 1);
					DBG("Alt key press: %s\n",gpio_buttons[i].desc);
				}
				else
				{
					input_report_key (kbd_data.idev, gpio_buttons[i].key, 1);
					DBG("Key press: %s\n",gpio_buttons[i].desc);
				}
			}
			else
			{
				kbd_data.backlight_timer = BACKLIGHT_TIMEOUT;
				gpio_set_value(EGPIO_1_2_KBD_BAKLIGHT,1);

				if (alternate_key)
				{
					input_report_key (kbd_data.idev, gpio_buttons[i].alt_key, 0);
					DBG("Alt key rel: %s\n",gpio_buttons[i].desc);
				}
				else
				{
					input_report_key (kbd_data.idev, gpio_buttons[i].key, 0);
					DBG("Key rel: %s\n",gpio_buttons[i].desc);
				}
				if (shift_key)
					input_report_key (kbd_data.idev, KEY_RIGHTSHIFT, 0);
				
				if (shift_key != PERMANENT)
					shift_key = FALSE;
				
				if (alternate_key != PERMANENT)
					alternate_key = FALSE;

			}
			input_sync(kbd_data.idev);
			kaiser_update_leds();
		}
	} 
	
	if (kbd_data.backlight_timer)
		kbd_data.backlight_timer--;
	else
		gpio_set_value(EGPIO_1_2_KBD_BAKLIGHT,0);
		gpio_set_value (18, 0);
		gpio_set_value (19, 0);
	spin_unlock_irqrestore (&kbd_data.lock, flags);

	kbd_data.timer20Hz.expires = jiffies+(HZ/20);
	add_timer(&kbd_data.timer20Hz);
}

static int kaiser_kbd_open (struct input_dev *idev)
{
	unsigned long   flags;
	int ret,i;
	DBG("Open!!!\n");
	kbd_data.workqueue = create_workqueue ("kaiserkbdw");
	INIT_WORK (&kbd_data.task, kaiser_kbd_queuework);

	//init timer to read gpio changes
	init_timer(&kbd_data.timer20Hz);
	kbd_data.timer20Hz.function = kbd_handle_timer20Hz;
	kbd_data.timer20Hz.expires = jiffies+(HZ/20);
	add_timer(&kbd_data.timer20Hz);

#ifdef USE_RELEASE_TIMER
	init_timer (&key_release_timer);
	key_release_timer.function = release_timer_went_off;
#endif

	// Set all GPIOS low
	spin_lock_irqsave (&kbd_data.lock,  flags);
	for (i = 0; i < MAX_ROW; i++)
		gpio_set_value (row_gpio[i], 0);

	spin_unlock_irqrestore (&kbd_data.lock,  flags);

	//ret = request_irq (INT_KEYSENSE, kaiser_kbd_irq_handler, IRQF_DISABLED, "kaiser-kbd",NULL);
	for (i = 0; i < ARRAY_SIZE(col_gpio);i++)
	{
		ret = request_irq (gpio_to_irq(col_gpio[i]), kaiser_kbd_irq_handler, IRQF_DISABLED, "kaiser-kbd",NULL);
		if (ret)
		{
			printk(KERN_ERR "Unable to register KBD IRQ!!!\n");
			return ret;
		}
		set_irq_type(gpio_to_irq(col_gpio[i]), IRQ_TYPE_EDGE_BOTH);
	}
	ret = request_irq (gpio_to_irq(18), kaiser_kbd_irq_handler, IRQF_DISABLED, "kaiser-kbd",NULL);
	ret = request_irq (gpio_to_irq(19), kaiser_kbd_irq_handler, IRQF_DISABLED, "kaiser-kbd",NULL);
	
	//set_irq_type(INT_KEYSENSE, IRQ_TYPE_EDGE_BOTH);
	return 0;
}

static void kaiser_kbd_close (struct input_dev *idev)
{
	int i;
	DBG("Close !!!\n");
	destroy_workqueue (kbd_data.workqueue);
	
	for (i=0;i<ARRAY_SIZE(col_gpio); i++)
		free_irq(gpio_to_irq(col_gpio[i]),NULL);
	//Jog wheel
	free_irq(gpio_to_irq(18),NULL);
	free_irq(gpio_to_irq(19),NULL);
	del_timer_sync(&kbd_data.timer20Hz);
#ifdef USE_RELEASE_TIMER
	del_timer (&key_release_timer);
#endif
	return;
}

static int __init kaiser_kbd_probe(struct platform_device *pdev)
{
	int             row, col;

	kbd_data.idev = input_allocate_device();
	kbd_data.idev->name = MODULE_NAME;
	kbd_data.idev->open = kaiser_kbd_open;
	kbd_data.idev->close = kaiser_kbd_close;
	kbd_data.idev->id.bustype = BUS_HOST;
	set_bit(EV_KEY, kbd_data.idev->evbit);
	set_bit(EV_REP, kbd_data.idev->evbit);
	for (row = 0; row < MAX_ROW; row++)
		for (col = 0; col < MAX_COL; col++)
		{
			if (matrix_buttons[row][col].key < KEY_MAX)
				set_bit(matrix_buttons[row][col].key, kbd_data.idev->keybit);

			if (matrix_buttons[row][col].alt_key < KEY_MAX)
				set_bit(matrix_buttons[row][col].alt_key, kbd_data.idev->keybit);;
		}

	for (col = 0; col < ARRAY_SIZE(gpio_buttons); col++)
	{
		if (gpio_buttons[col].key < KEY_MAX)
			set_bit(gpio_buttons[col].key, kbd_data.idev->keybit);
		if (gpio_buttons[col].alt_key < KEY_MAX)
			set_bit(gpio_buttons[col].alt_key, kbd_data.idev->keybit);
	}

	if (input_register_device (kbd_data.idev))
	{
		input_free_device(kbd_data.idev);
		return -ENODEV;
	}

	return 0;
}

static int __init kaiser_kbd_remove(struct platform_device *pdev)
{
	DBG ("removing device...\n");
	input_unregister_device (kbd_data.idev);
	return 0;
}

static struct platform_driver kaiser_kbd_driver = {
	.probe = kaiser_kbd_probe,
  .remove = kaiser_kbd_remove,
	.driver = {
		.name = MODULE_NAME,
	},
};

static int __init kaiser_kbd_init (void)
{
	int ret;

	ret = platform_driver_register(&kaiser_kbd_driver);
	if (ret)
	{
		printk(KERN_ERR "Cant register platform_driver:" MODULE_NAME);
		return ret;
	}

	kbd_data.platform_dev = platform_device_register_simple(MODULE_NAME, -1, NULL, 0);

	if (IS_ERR(kbd_data.platform_dev))
	{
		platform_driver_unregister(&kaiser_kbd_driver);
		printk(KERN_ERR "Cant register device " MODULE_NAME);
		return PTR_ERR(kbd_data.platform_dev);;
	}
	return ret;
}

static void __exit kaiser_kbd_cleanup (void)
{
	platform_device_unregister(kbd_data.platform_dev);
  platform_driver_unregister(&kaiser_kbd_driver);
}

module_init (kaiser_kbd_init);
module_exit (kaiser_kbd_cleanup);

MODULE_AUTHOR ("Marcin Baliniak");
MODULE_DESCRIPTION ("Support for HTC Kaiser Keyboard");
MODULE_LICENSE ("GPL");
