/* arch/arm/mach-msm/vogue-smd.c
 * Author: Martin Johnson <M.J.Johnson@massey.ac.nz>
 * Based on smd.c by Brian Swetland <swetland@google.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/platform_device.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/debugfs.h>
#include <linux/delay.h>
#include <asm/arch/msm_smd.h>
#include <asm/arch/msm_iomap.h>
#include <asm/arch/system.h>
#include <asm/io.h>

#include "smd_private.h"
#include "proc_comm.h"

#define MODULE_NAME "msm_smd"

#define BURST_SIZE 0x10000

void *smem_find(unsigned id, unsigned size);
void smd_diag(void);

#if 0
#define D(x...) printk(x)
#else
#define D(x...) do {} while (0)
#endif

#define MSM_A2M_INT(n) (MSM_CSR_BASE + 0x400 + (n) * 4)


static inline void notify_other_smd(int ch)
{
	//	printk("[n:%d] ",ch);
	writel(1, MSM_A2M_INT(3+ch));
}


// not working yet
int msm_proc_comm(unsigned cmd, unsigned *data1, unsigned *data2)
{
	return -1;
}

/* the spinlock is used to synchronize between the
** irq handler and code that mutates the channel
** list or fiddles with channel state
*/
static DEFINE_SPINLOCK(smd_lock);

/* the mutex is used during open() and close()
** operations to avoid races while creating or
** destroying smd_channel structures
*/
static DEFINE_MUTEX(smd_creation_mutex);

static int smd_initialized = 0;


struct smd_buffer
{
// these are stored in shared memory and used to communicate with the ARM9
	volatile unsigned short head;
	volatile unsigned short tail;
} smd_buffer_t;


// fifo in shared memory
struct smd_half_channel
{
	// pointer to heand a tail of the fifo
	volatile struct smd_buffer *buffer;
	// start and size in shared memory of a single fifo
	unsigned size;
	char *start;
};

static LIST_HEAD(smd_ch_list);


struct smd_channel
{
	// a channel has 2 fifos.
	volatile struct smd_half_channel send;
	volatile struct smd_half_channel recv;
	struct list_head ch_list;

	//	unsigned current_packet;
	unsigned n;
	void *priv;
	void (*notify)(void *priv, unsigned flags);
	/*
	int (*read)(smd_channel_t *ch, void *data, int len);
	int (*write)(smd_channel_t *ch, const void *data, int len);
	int (*read_avail)(smd_channel_t *ch);
	int (*write_avail)(smd_channel_t *ch);

	void (*update_state)(smd_channel_t *ch);
	unsigned last_state;
	*/
};

/* how many bytes are available for reading */
int smd_read_avail(struct smd_channel *ch)
{
	return (ch->recv.buffer->head - ch->recv.buffer->tail + ch->recv.size) % ch->recv.size;
}

/* how many bytes we are free to write */
int smd_write_avail(struct smd_channel *ch)
{
	return (ch->send.size - 1) -
		((ch->send.buffer->head - ch->send.buffer->tail + ch->send.size) % ch->send.size);
}


void check_for_smd_data(void) {
	struct smd_channel *ch;
	unsigned long flags;
	
	spin_lock_irqsave(&smd_lock, flags);
	list_for_each_entry(ch, &smd_ch_list, ch_list) {
		if(ch->recv.buffer->head != ch->recv.buffer->tail)
			ch->notify(ch->priv, SMD_EVENT_DATA);
		if(ch->send.buffer->head != ch->send.buffer->tail)			notify_other_smd(ch->n);
	}
	spin_unlock_irqrestore(&smd_lock, flags);
}

static irqreturn_t smd_irq_handler(int irq, void *data)
{
	//unsigned long flags;
	//	printk("[%d] ",irq);
	check_for_smd_data();
	return IRQ_HANDLED;
}

void smd_kick(smd_channel_t *ch)
{
	unsigned long flags;

	spin_lock_irqsave(&smd_lock, flags);
	//	ch->update_state(ch);
	ch->notify(ch->priv, SMD_EVENT_DATA);
	notify_other_smd(ch->n);
	spin_unlock_irqrestore(&smd_lock, flags);
}

// write to smd fifo, notify after
int smd_write(smd_channel_t *ch, const void *_data, int len)
{
	const unsigned char *buf = _data;
	int sent = 0;
	unsigned short myhead;
	check_for_smd_data();

	D("smd_stream_write() %d -> ch%d\n", len, ch->n);
	if (len < 0) return -EINVAL;

	myhead=ch->send.buffer->head;
	notify_other_smd(ch->n);
	while(len!=0) {
		myhead=(myhead+1)%ch->send.size;
		len=len-1;
	 	*(ch->send.start+myhead)=*buf++;
		sent++;
		
		if((sent%BURST_SIZE)==0) {
			ch->send.buffer->head=myhead;
			notify_other_smd(ch->n);
		}
	
	}
	ch->send.buffer->head=myhead;

	notify_other_smd(ch->n);
	
	check_for_smd_data();
	return sent;
}

// read data from smd fifo, it seems to like reading in bursts of 0xf4 bytes and then notifying the ARM9
int smd_read(smd_channel_t *ch, void *data, int len)
{
	int recvd=0;
	unsigned short mytail;
	unsigned char *buf = data;
	
	if (len < 0) return -EINVAL;
	notify_other_smd(ch->n);	
	mytail=ch->recv.buffer->tail;
	while(recvd!=len) {
		mytail=(mytail+1)%ch->recv.size;
		if(buf)
			*buf++=*(ch->recv.start+mytail);
		recvd++;
		if((recvd%BURST_SIZE)==0) {
			ch->recv.buffer->tail=mytail;
			notify_other_smd(ch->n);
		}
	}
	ch->recv.buffer->tail=mytail;
	//	printk("{r:%d<-%d] ",recvd,ch->n);
	notify_other_smd(ch->n);
	
	return recvd;
}


static void do_nothing_notify(void *priv, unsigned flags)
{
}


int smd_open(int n, smd_channel_t **_ch, void *priv, void (*notify)(void *, unsigned))
{
	struct smd_channel *ch;
	unsigned long flags;
	int res = 0;

	if (smd_initialized == 0) {
		printk(KERN_INFO "smd_open() before smd_init()\n");
		return -ENODEV;
	}

	if(n!=0 && n!=1)
		return -ENODEV;

	if (notify == 0)
		notify = do_nothing_notify;


	mutex_lock(&smd_creation_mutex);

        spin_lock_irqsave(&smd_lock, flags);
        list_for_each_entry(ch, &smd_ch_list, ch_list) {
                if (ch->n == n) {
                        res = -ENODEV;
                        break;
                }
        }
        spin_unlock_irqrestore(&smd_lock, flags);
        if (res) goto out;


	ch = kzalloc(sizeof(struct smd_channel), GFP_KERNEL);
	if (ch == 0) {
		res = -ENOMEM;
		goto out;
	}
/*
all offsets from SMEM_BASE
circular buffer at 0xf0000 0x200 in length for AT receive, head and tail at 0xf3fc8
circular buffer at 0xf0200 0x200 in length for AT send, head and tail at 0xf3fcc

circular buffer at 0xf0a00 0xd82 in length for SMD1 receive, head and tail at 0xf3fd8
circular buffer at 0xf1782 0x282 in length for SMD1 send, head and tail at 0xf3fdc
*/
	if(n==0) {
		ch->send.buffer=(struct smd_buffer *)(MSM_SHARED_RAM_BASE+0xf3fcc);
		ch->send.start=(char *)(MSM_SHARED_RAM_BASE+0xf0200);
		ch->send.size=0x200;
		ch->recv.buffer=(struct smd_buffer *)(MSM_SHARED_RAM_BASE+0xf3fc8);
		ch->recv.start=(char *)(MSM_SHARED_RAM_BASE+0xf0000);
		ch->recv.size=0x200;
	} else {
		ch->send.buffer=(struct smd_buffer *)(MSM_SHARED_RAM_BASE+0xf3fdc);
		ch->send.start=(char *)(MSM_SHARED_RAM_BASE+0xf1782);
		ch->send.size=0x282;
		ch->recv.buffer=(struct smd_buffer *)(MSM_SHARED_RAM_BASE+0xf3fd8);
		ch->recv.start=(char *)(MSM_SHARED_RAM_BASE+0xf0a00);
		ch->recv.size=0xd82;
	}
	ch->priv = priv;
	ch->notify = notify;
	ch->n = n;
	/*
	ch->current_packet = 0;
	ch->read = smd_stream_read;
	ch->write = smd_stream_write;
	ch->read_avail = smd_stream_read_avail;
	ch->write_avail = smd_stream_write_avail;
	ch->update_state = update_stream_state;
*/
	*_ch = ch;

	spin_lock_irqsave(&smd_lock, flags);
	list_add(&ch->ch_list, &smd_ch_list);
	spin_unlock_irqrestore(&smd_lock, flags);
	smd_kick(ch);

out:
	mutex_unlock(&smd_creation_mutex);
	return res;
}

int smd_close(smd_channel_t *ch)
{
	unsigned long flags;

	if (ch == 0) return -1;

	mutex_lock(&smd_creation_mutex);
	spin_lock_irqsave(&smd_lock, flags);

	list_del(&ch->ch_list);
	D("smd_close: closing ch %d\n", ch->n);

	spin_unlock_irqrestore(&smd_lock, flags);
	mutex_unlock(&smd_creation_mutex);

	return 0;
}
/*
int smd_read(smd_channel_t *ch, void *data, int len)
{
	return ch->read(ch, data, len);
}

int smd_write(smd_channel_t *ch, const void *data, int len)
{
	return ch->write(ch, data, len);
}

int smd_read_avail(smd_channel_t *ch)
{
	return ch->read_avail(ch);
}

int smd_write_avail(smd_channel_t *ch)
{
	return ch->write_avail(ch);
}

int smd_wait_until_readable(smd_channel_t *ch, int bytes)
{
	return -1;
}

int smd_wait_until_writable(smd_channel_t *ch, int bytes)
{
	return -1;
}

int smd_cur_packet_size(smd_channel_t *ch)
{
	return ch->current_packet;
}
*/

/* --------------------------------------------------------------------------------------- */

void *smem_alloc(unsigned id, unsigned size)
{
	return smem_find(id, size);
}

void *smem_find(unsigned id, unsigned size_in)
{
	return 0;
}

int smd_core_init(void)
{
	int r;
	int i;
	printk(KERN_INFO "smd_core_init()\n");

	// these interrupts are used by the ARM9 to tell us that data is ready to read.
	for(i=0;i<7;i++) {
		r = request_irq(INT_A9_M2A_0+i, smd_irq_handler,
			IRQF_TRIGGER_RISING, "smd_dev", 0);
		if (r < 0)
			return r;
		r = enable_irq_wake(INT_A9_M2A_0+i);
		if (r < 0)
			printk(KERN_ERR "smd_core_init: enable_irq_wake failed for INT_A9_M2A_%d\n",i);
	}

	printk(KERN_INFO "smd_core_init() done\n");

	return 0;
}


static int __init msm_smd_probe(struct platform_device *pdev)
{
	printk(KERN_INFO "smd_init()\n");

	if (smd_core_init()) {
		printk(KERN_ERR "smd_core_init() failed\n");
		return -1;
	}

	smd_initialized = 1;

	return 0;
}

static struct platform_driver msm_smd_driver = {
	.probe = msm_smd_probe,
	.driver = {
		.name = MODULE_NAME,
		.owner = THIS_MODULE,
	},
};

static int __init msm_smd_init(void)
{
	return platform_driver_register(&msm_smd_driver);
}

module_init(msm_smd_init);

MODULE_DESCRIPTION("MSM Shared Memory Core for Vogue");
MODULE_AUTHOR("Martin Johnson <>");
MODULE_LICENSE("GPL");
